Attribute VB_Name = "modTrickMemReader"
' // modTrickMemReader.bas  -  
' //     (The trick), 2014

Option Explicit

Public Enum NT_STATUS
    STATUS_SUCCESS = 0
    STATUS_INVALID_PARAMETER = &HC000000D
End Enum

Public Type UNICODE_STRING
    Length              As Integer
    MaximumLength       As Integer
    lpBuffer            As Long
End Type

Public Type LIST_ENTRY
    Flink               As Long
    Blink               As Long
End Type

Public Type KDEVICE_QUEUE
    Type                As Integer
    Size                As Integer
    DeviceListHead      As LIST_ENTRY
    Lock                As Long
    Busy                As Long
End Type

Public Type KDPC
    Type                As Byte
    Importance          As Byte
    Number              As Integer
    DpcListEntry        As LIST_ENTRY
    DeferredRoutine     As Long
    DeferredContext     As Long
    SystemArgument1     As Long
    SystemArgument2     As Long
    DpcData             As Long
End Type

Public Type DISPATCHER_HEADER
    Lock                As Long
    SignalState         As Long
    WaitListHead        As LIST_ENTRY
End Type

Public Type KEVENT
    Header              As DISPATCHER_HEADER
End Type

Public Type IO_STATUS_BLOCK
    StatusPointer       As Long
    Information         As Long
End Type

Public Type Tail
    DriverContext(3)    As Long
    Thread              As Long
    AuxiliaryBuffer     As Long
    ListEntry           As LIST_ENTRY
    lpCurStackLocation  As Long
    OriginalFileObject  As Long
End Type

Public Type IRP
    Type                As Integer
    Size                As Integer
    MdlAddress          As Long
    Flags               As Long
    AssociatedIrp       As Long
    ThreadListEntry     As LIST_ENTRY
    IoStatus            As IO_STATUS_BLOCK
    RequestorMode       As Byte
    PendingReturned     As Byte
    StackCount          As Byte
    CurrentLocation     As Byte
    Cancel              As Byte
    CancelIrql          As Byte
    ApcEnvironment      As Byte
    AllocationFlags     As Byte
    UserIosb            As Long
    UserEvent           As Long
    Overlay             As Currency
    CancelRoutine       As Long
    UserBuffer          As Long
    Tail                As Tail
End Type

Public Type DEVICEIOCTL
    OutputBufferLength  As Long
    InputBufferLength   As Long
    IoControlCode       As Long
    Type3InputBuffer    As Long
End Type

Public Type IO_STACK_LOCATION
    MajorFunction       As Byte
    MinorFunction       As Byte
    Flags               As Byte
    Control             As Byte
    ' //  DeviceIoControl  
    DeviceIoControl     As DEVICEIOCTL
    pDeviceObject       As Long
    pFileObject         As Long
    pCompletionRoutine  As Long
    pContext            As Long
End Type

Public Type DRIVER_OBJECT
    Type                As Integer
    Size                As Integer
    pDeviceObject       As Long
    Flags               As Long
    DriverStart         As Long
    DriverSize          As Long
    DriverSection       As Long
    DriverExtension     As Long
    DriverName          As UNICODE_STRING
    HardwareDatabase    As Long
    FastIoDispatch      As Long
    DriverInit          As Long
    DriverStartIo       As Long
    DriverUnload        As Long
    MajorFunction(27)   As Long
End Type

Public Type DEVICE_OBJECT
    Type                As Integer
    Size                As Integer
    ReferenceCount      As Long
    DriverObject        As Long
    NextDevice          As Long
    AttachedDevice      As Long
    CurrentIrp          As Long
    Timer               As Long
    Flags               As Long
    Characteristics     As Long
    Vpb                 As Long
    DeviceExtension     As Long
    DeviceType          As Long
    StackSize           As Byte
    Queue(39)           As Byte
    AlignRequirement    As Long
    DeviceQueue         As KDEVICE_QUEUE
    Dpc                 As KDPC
    ActiveThreadCount   As Long
    SecurityDescriptor  As Long
    DeviceLock          As KEVENT
    SectorSize          As Integer
    Spare1              As Integer
    DeviceObjExtension  As Long
    Reserved            As Long
End Type
Private Type BinaryString
    D(255)              As Integer
End Type

Public Const FILE_DEVICE_UNKNOWN    As Long = &H22
Public Const IO_NO_INCREMENT        As Long = &H0
Public Const IRP_MJ_CREATE          As Long = &H0
Public Const IRP_MJ_CLOSE           As Long = &H2
Public Const IRP_MJ_DEVICE_CONTROL  As Long = &HE
Public Const FILE_DEVICE_MEMREADER  As Long = &H8000&
Public Const IOCTL_READ_MEMORY      As Long = &H80002000

Public DeviceName       As UNICODE_STRING   ' //    
Public DeviceLink       As UNICODE_STRING   ' //    
Public Device           As DEVICE_OBJECT    ' //  

Dim strName As BinaryString     ' //    
Dim strLink As BinaryString     ' //    

Public Sub Main()
End Sub

' //   - False
Public Function NT_SUCCESS( _
                ByVal Status As NT_STATUS) As Boolean
    NT_SUCCESS = Status >= STATUS_SUCCESS
End Function

' //     
Public Function IoGetCurrentIrpStackLocation( _
                ByRef pIrp As IRP) As Long
    IoGetCurrentIrpStackLocation = pIrp.Tail.lpCurStackLocation
End Function

' //    
Public Function DriverEntry( _
                ByRef DriverObject As DRIVER_OBJECT, _
                ByRef RegistryPath As UNICODE_STRING) As NT_STATUS
    Dim Status As NT_STATUS
    
    ' //  
    Status = Init()
    
    ' //    ,   , ..    Init
    If Not NT_SUCCESS(Status) Then
    
        DriverEntry = Status
        Exit Function
        
    End If
    
    ' //  
    Status = IoCreateDevice(DriverObject, 0, DeviceName, FILE_DEVICE_MEMREADER, 0, False, Device)
    
    ' //    
    If Not NT_SUCCESS(Status) Then
    
        DriverEntry = Status
        Exit Function
        
    End If
    
    ' //         
    Status = IoCreateSymbolicLink(DeviceLink, DeviceName)
    
    ' //  
    If Not NT_SUCCESS(Status) Then
    
        ' //    
        IoDeleteDevice Device
        DriverEntry = Status
        Exit Function
        
    End If
    
    ' //  
    DriverObject.DriverUnload = GetAddr(AddressOf DriverUnload)                                 ' //  
    DriverObject.MajorFunction(IRP_MJ_CREATE) = GetAddr(AddressOf DriverCreateClose)            ' //   CreateFile
    DriverObject.MajorFunction(IRP_MJ_CLOSE) = GetAddr(AddressOf DriverCreateClose)             ' //   CloseHandle
    DriverObject.MajorFunction(IRP_MJ_DEVICE_CONTROL) = GetAddr(AddressOf DriverDeviceControl)  ' //   DeviceIoControl
    
    ' // 
    DriverEntry = STATUS_SUCCESS
    
End Function

' //   
Public Sub DriverUnload( _
           ByRef DriverObject As DRIVER_OBJECT)
           
    ' //  
    IoDeleteSymbolicLink DeviceLink
    ' //  
    IoDeleteDevice ByVal DriverObject.pDeviceObject
    
End Sub

' //    / 
Public Function DriverCreateClose( _
                ByRef DeviceObject As DEVICE_OBJECT, _
                ByRef pIrp As IRP) As NT_STATUS
                
    pIrp.IoStatus.Information = 0
    pIrp.IoStatus.StatusPointer = STATUS_SUCCESS
    
    ' //  IRP   /
    IoCompleteRequest pIrp, IO_NO_INCREMENT
    
    ' // 
    DriverCreateClose = STATUS_SUCCESS
    
End Function

' //   IOCTL 
Public Function DriverDeviceControl( _
                ByRef DeviceObject As DEVICE_OBJECT, _
                ByRef pIrp As IRP) As NT_STATUS
    Dim lpStack As Long
    Dim ioStack As IO_STACK_LOCATION
    
    ' //     
    lpStack = IoGetCurrentIrpStackLocation(pIrp)
    
    ' //    
    If lpStack Then
    
        ' //    
        memcpy ioStack, ByVal lpStack, Len(ioStack)
        
        ' //  IOCTL   AssociatedIrp    SystemBuffer
        ' //  SystemBuffer  ,    DeviceIoControl
        If ioStack.DeviceIoControl.IoControlCode = IOCTL_READ_MEMORY And _
            pIrp.AssociatedIrp <> 0 Then

            Dim lpPointer   As Long
            Dim DataSize    As Long
            
            ' //    SystemBuffer
            memcpy lpPointer, ByVal pIrp.AssociatedIrp, 4
            memcpy DataSize, ByVal pIrp.AssociatedIrp + 4, 4
            
            ' //   
            If DataSize <= ioStack.DeviceIoControl.OutputBufferLength Then
            
                ' //   ,    
                Dim lpStart As Long
                Dim pgCount As Long
                Dim pgSize  As Long
                Dim pgOfst  As Long
                
                ' //    
                lpStart = lpPointer And &HFFFFF000
                
                ' //     
                pgOfst = lpPointer And &HFFF&
                
                ' //       PageFault
                Do While MmIsAddressValid(ByVal lpStart) And (pgSize - pgOfst < DataSize)
                
                    lpStart = lpStart + &H1000
                    pgCount = pgCount + 1
                    pgSize = pgSize + &H1000
                    
                Loop
                
                ' //     
                If pgCount Then
                
                    ' //     
                    pgSize = pgCount * &H1000 - pgOfst
                    
                    ' //  
                    If DataSize > pgSize Then DataSize = pgSize
                    
                    ' //     
                    pIrp.IoStatus.Information = DataSize
                    
                    ' // 
                    pIrp.IoStatus.StatusPointer = STATUS_SUCCESS
                    
                    ' //    SystemBuffer
                    memcpy ByVal pIrp.AssociatedIrp, ByVal lpPointer, DataSize
                    
                    ' //  IRP   /
                    IoCompleteRequest pIrp, IO_NO_INCREMENT
                    
                    ' // 
                    DriverDeviceControl = STATUS_SUCCESS
                    
                    ' // 
                    Exit Function
    
                End If
                
            End If
    
        End If
        
    End If
    
    ' //     
    pIrp.IoStatus.Information = 0
    
    ' //  DeviceIoControl
    pIrp.IoStatus.StatusPointer = STATUS_INVALID_PARAMETER
    
    ' //  IRP   /
    IoCompleteRequest pIrp, IO_NO_INCREMENT
    
    ' // 
    DriverDeviceControl = STATUS_INVALID_PARAMETER
    
End Function

' //  
Private Function Init() As NT_STATUS

    ' //   
    ' //\Device\TrickMemReader
    strName.D(0) = &H5C:    strName.D(1) = &H44:    strName.D(2) = &H65:    strName.D(3) = &H76:    strName.D(4) = &H69:
    strName.D(5) = &H63:    strName.D(6) = &H65:    strName.D(7) = &H5C:    strName.D(8) = &H54:    strName.D(9) = &H72:
    strName.D(10) = &H69:   strName.D(11) = &H63:   strName.D(12) = &H6B:   strName.D(13) = &H4D:   strName.D(14) = &H65:
    strName.D(15) = &H6D:   strName.D(16) = &H52:   strName.D(17) = &H65:   strName.D(18) = &H61:   strName.D(19) = &H64:
    strName.D(20) = &H65:   strName.D(21) = &H72:
    
    ' //  UNICODE_STRING
    RtlInitUnicodeString DeviceName, strName
    
    ' //       user-mode
    ' //\DosDevices\TrickMemReader
    strLink.D(0) = &H5C:    strLink.D(1) = &H44:    strLink.D(2) = &H6F:    strLink.D(3) = &H73:    strLink.D(4) = &H44:
    strLink.D(5) = &H65:    strLink.D(6) = &H76:    strLink.D(7) = &H69:    strLink.D(8) = &H63:    strLink.D(9) = &H65:
    strLink.D(10) = &H73:   strLink.D(11) = &H5C:   strLink.D(12) = &H54:   strLink.D(13) = &H72:   strLink.D(14) = &H69:
    strLink.D(15) = &H63:   strLink.D(16) = &H6B:   strLink.D(17) = &H4D:   strLink.D(18) = &H65:   strLink.D(19) = &H6D:
    strLink.D(20) = &H52:   strLink.D(21) = &H65:   strLink.D(22) = &H61:   strLink.D(23) = &H64:   strLink.D(24) = &H65:
    strLink.D(25) = &H72:
    
    ' //  UNICODE_STRING
    RtlInitUnicodeString DeviceLink, strLink

End Function

Private Function GetAddr( _
                 ByVal Value As Long) As Long
    GetAddr = Value
End Function
